from pebble import ProcessPool
from concurrent.futures import TimeoutError
import time
import os
import sys
import platform
import subprocess
import logging
import warnings
import re
import requests
import socket
import json
try:
    import uptime
except:
    pass
import datetime

try:
    import psutil
except:
    pass

try:
    import distro
except:
    pass

class DataGatherer():

    def __init__(self, clientVersion = ''):
        self.clientVersion = clientVersion

    data = {}

    def does_exist(self, name):
        """Check whether `name` is on PATH and marked as executable."""
        try:
            # from whichcraft import which
            from shutil import which

            return which(name) is not None
        except:
            return False

    def worker(self, task):
        try:
            method = getattr(self, task)
            return [task, method()]
        except AttributeError:
            print('No such function ' + task)

    def callback(self, future):
        try:
            result = future.result()  # blocks until results are ready
            if result:
                self.data[result[0]] = result[1]
        except TimeoutError as error:
            print('Function took longer than %d seconds' % error.args[1])
        #xcept Exception as error:
            ##print('Function raised %s' % error)
            ##print(error.traceback)  # traceback of the function

    def gather(self, tasks, timeout):
        with ProcessPool(max_workers=5, max_tasks=20) as pool:
            for task in tasks:
                future = pool.schedule(self.worker, args=[task], timeout=timeout)
                future.add_done_callback(self.callback)
#max tasks increased from 10 to 20 to accomodate the get_jobs task -LVT
    def terminalCommand(self, cmd, multipleLines = False):
        '''Strips extraneous characters from terminal output.'''
        with subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT) as p:
            if multipleLines:
                return str(p.stdout.readlines())
            return str(p.stdout.readline().decode('utf-8').strip('\n'))

    def hostname(self):
        '''Returns the hostname of the host. Cross platform.'''
        return os.getenv('HOSTNAME', os.getenv('COMPUTERNAME', platform.node())).split('.')[0]

    def name(self):
        '''Returns the hostname of the host. Cross platform.'''
        return self.hostname()

    def load_average(self):
        '''Returns the load average of the host'''
        try:
            avg =  os.getloadavg()
            avgString = ', '.join([str(x) for x in avg])
            return avgString
        except (ValueError, OSError):
            return ''

    def uptime(self):
        return uptime.uptime()

    def last_sw_update(self):
        try:
            lastUpdate = self.terminalCommand('stat -c %Y /var/cache/apt/', False)
            lastUpdateDateTime = datetime.datetime.fromtimestamp(float(lastUpdate)).strftime('%Y-%m-%d %H:%M:%S')
        except:
            lastUpdateDateTime = ''
        return lastUpdateDateTime

    def bootTime(self):
        '''Returns boot time of host'''
        return uptime.boottime()

    def pythonVersion(self):
        version = str(sys.version_info[0]) + '.' + str(sys.version_info[1]) + str(sys.version_info[2])
        return version

    def ram(self):
        # returns memory in bytes
        return os.sysconf('SC_PAGE_SIZE') * os.sysconf('SC_PHYS_PAGES')  # e.g. 4015976448

    def free_ram(self):
        try:
            # returns memory in bytes
            mem = psutil.virtual_memory()
            return mem.available  # e.g. 4015976448
        except:
            return ''

    def cpu(self):
        if platform.system() == 'Windows':
            return platform.processor()
        elif platform.system() == 'Darwin':
            return self.terminalCommand('sysctl -n machdep.cpu.brand_string')
        elif platform.system() == 'Linux':
            command = 'cat /proc/cpuinfo'
            all_info = subprocess.check_output(command, shell=True).strip()
            for line in all_info.decode('utf-8').split('\n'):
                if 'model name' in line:
                    return re.sub( '.*model name.*:', '', line,1)
        return ''

    def ipv4_address(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            # doesn't even have to be reachable
            s.connect(('10.255.255.255', 1))
            IP = s.getsockname()[0]
        except:
            IP = '127.0.0.1'
        finally:
            s.close()
        return IP

    def ipv6_address(self):
        return ''
        
        

    def primary_drive(self):
        drive = None
        try:
            stat= os.statvfs('/')
            drive = '/'
        except:
            pass
        return drive
            
    def primary_drive_size(self):
        total = None
        try:
            stat= os.statvfs('/')
            total = (stat.f_blocks * stat.f_frsize)
        except:
            pass
        return total
    
    def primary_drive_free_space(self):
        free = None
        try:
            stat= os.statvfs('/')
            free = (stat.f_bavail * stat.f_frsize)
        except:
            pass
        return free

    def wan_ipv4_address(self):
        try:
            return requests.get('https://api.ipify.org', timeout=8).text
        except:
            return ''

    def wan_ipv6_address(self):
        return ''

    def os_version(self):
        try: 
        
            os_type = sys.platform.lower()
            if "darwin" in os_type:
                return 'macOS ' + platform.mac_ver()[0]
            elif "linux" in os_type:
                try:
                    return distro.id() + ' ' + distro.version()
                except:
                    return platform.linux_distribution()[0] + ' ' + platform.linux_distribution()[1]
            elif "win" in os_type:
                return platform.system() + ' ' + platform.release()
        
        except:
            return ''
    
    def os_name(self):
        try: 
        
            os_type = sys.platform.lower()
            if "linux" in os_type:
                try:
                    return distro.id()
                except:
                    return platform.linux_distribution()[0]
            elif "win" in os_type:
                return platform.system()
        
        except:
            return ''

    def os_build(self):
        try: 
        
            os_type = sys.platform.lower()
            if "linux" in os_type:
                try:
                    return distro.version()
                except:
                    return platform.linux_distribution()[1]
            elif "win" in os_type:
                return platform.release()
        
        except:
            return ''
    
    def agent_version(self):
        return self.clientVersion

    def agent_type(self):
        return 'python3'

    def get_jobs(self):
        return 'false'

    def job_results(self):
        return []

    def hw_serial_number(self):
        try:
            os_type = sys.platform.lower()
            if "darwin" in os_type:
                command = "ioreg -l | awk '/IOPlatformSerialNumber/ { print $4;}'"
            elif "linux" in os_type:
                if (self.does_exist('hal-get-property')):
                    command = "hal-get-property --udi /org/freedesktop/Hal/devices/computer --key system.hardware.uuid"
                elif (self.does_exist('dmidecode')):
                    command = "dmidecode -t system | grep Serial"
                else:
                    command = ''

            elif "win" in os_type:
                command = "wmic bios get serialnumber"
            
            if (command != ''):
                return os.popen(command).read().replace("\n","").replace("	","").replace(" ","").replace("SerialNumber:","").replace("SerialNumber","").replace("\"","")
            else:
                return ''
        except:
            return ''